/***************************************************************************
 *                                                                         *
 *   PLIS COMPILER                                                         *
 *   =============                                                         *
 *                                                                         *
 *   Copyright (C) Juan Pedro Bolivar Puente 2007                          *
 *                                                                         *
 *   This program is free software: you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 *                                                                         *
 ***************************************************************************/

#include <fstream>

#include "compiler.h"
#include "misc.h"

using namespace std;

const char* REG_NAME[MAX_REG][TYPE_MAX_SIZE] = {
    {"REG_ERROR", "REG_ERROR", "REG_ERROR", "REG_ERROR", "REG_ERROR"},
    {"REG_ERROR", "%al", "%ax", "REG_ERROR", "%eax"},
    {"REG_ERROR", "%dl", "%dx", "REG_ERROR", "%edx"},
    {"REG_ERROR", "%bl", "%bx", "REG_ERROR", "%ebx"},
    {"REG_ERROR", "%cl", "%cx", "REG_ERROR", "%ecx"}
};

const char SIZE_MOD[TYPE_MAX_SIZE] = {
    'X', 'b', 'w', 'X', 'l'
};


std::ostream& Compiler::sink()
{
    return **(--m_code_blocks.end());
}

std::string Compiler::labelizeString(const std::string& s)
{
    if (s.empty())
	return s;
    
    size_t i;
    ostringstream out;

    if (isdigit(s[0]))
	out << "__";
    
    for (i = 0; i < s.size(); ++i) {
	if (isalpha(s[i]) || isdigit(s[i]) || s[i] == '_')
	    out << s[i];
	else
	    out << "__" << (int)s[i] << "__";
    }

    return out.str();
}

std::string Compiler::getNewLabel()
{
    return string(".L") + itoa(m_label_count++,16);
}

Jump Compiler::getNewJump(bool follows)
{
    Jump j;
    j.name = getNewLabel();
    j.follows = follows;
    return j;
}

void Compiler::structIf()
{
    advance();
    
    Jump jmpif = getNewJump(true);
    Jump jmpelse = getNewJump(false);

    conditional(jmpif, jmpelse);

    sink() << jmpif.name << ":\n";
    while(expandSentence());

    sink() << jmpelse.name << ":\n";
    if (m_token->getType() == Token::STRUCT_ELSE) {
	advance();
	while(expandSentence());
    }
}

void Compiler::structWhile()
{
    advance();
    
    Jump loop_start = getNewJump(true);
    Jump loop_end = getNewJump(false);
    string loop_cond = getNewLabel();

    m_loop_tag = loop_cond;
    m_stop_tag = loop_end.name;
    
    sink() << loop_cond << ":\n";

    conditional(loop_start, loop_end);
    
    sink() << loop_start.name << ":\n";

    while(expandSentence());

    sink() << "\tjmp " << loop_cond << "\n";
    sink() << loop_end.name << ":\n";

    m_loop_tag.clear();
    m_stop_tag.clear();
}

void Compiler::structRepeat()
{
    advance();

    Jump loop_start = getNewJump(false);
    Jump loop_end = getNewJump(true);

    m_loop_tag = loop_start.name;
    m_stop_tag = loop_end.name;
    
    sink() << loop_start.name << ":\n";

    while(expandSentence());
    conditional(loop_start, loop_end);

    sink() << loop_end.name << ":\n";

    m_loop_tag.clear();
    m_stop_tag.clear();
}

void Compiler::structFor()
{
    throw Exception(Exception::UNIMPLEMENTED_FEATURE, *m_token);
}

void Compiler::expandReturn()
{
    advance();
    
    if (m_ret_type.type != Type::NONE) {
	Value* val = expression(true);

	if (val->reg != REG_EAX)
	    sink() << "\n\tmov" << SIZE_MOD[val->size()] << ' '
		   << outputValue(val) << ','
		   << REG_NAME[REG_EAX][val->size()] << endl;

	clear(val);
	delete val;
    }
    
    m_code_blocks.push_back(new ostringstream);
}

void Compiler::expandFlowStop()
{
    advance();
    
    if (m_loop_tag.empty())
	throw Exception(Exception::FLOW_MODIFIER_OUTSIDE_LOOP);

    sink() << "\tjmp " << m_stop_tag << endl;
}

void Compiler::expandFlowLoop()
{
    advance();
    
    if (m_stop_tag.empty())
	throw Exception(Exception::FLOW_MODIFIER_OUTSIDE_LOOP);

    sink() << "\tjmp " << m_loop_tag << endl;
}

bool Compiler::expandSentence()
{
    switch(m_token->getType()) {
    case Token::VOID_SENTENCE:
	advance();
	sink() << "\n"; /* A trick to add a new line in the source */
	break;
    case Token::END_BLOCK:
	advance();
	return false;
    case Token::STRUCT_IF:
	structIf();
	break;
    case Token::STRUCT_WHILE:
	structWhile();
	break;
    case Token::STRUCT_REPEAT:
	structRepeat();
	break;
    case Token::STRUCT_FOR:
	structFor();
	break;
    case Token::RETURN:
	expandReturn();
	break;
    case Token::FLOW_LOOP:
	expandFlowLoop();
	break;
	
    case Token::FLOW_STOP:
	expandFlowStop();
	break;
	
    default:
	delete expression(false);
	resetRegisters();
	resetTemporals();
	m_tmp_str.clear();
	break;
    }

    return true;
}

void Compiler::readType(Type& t, bool allow_none)
{
    if (m_token->getType() == Token::TYPE_CONST) {
	t.con = true; /* TODO */
	advance();
    } else
	t.con = false;
     
    while (m_token->getType() == Token::TYPE_PTR) {
	t.ptr++;
	advance();
    }

    if (m_token->getType() == Token::TYPE_UNSIGNED) {
	t.unsign = true;
	advance();
    } else
	t.unsign = false;
    
    switch (m_token->getType()) {
    case Token::TYPE_INT :
	t.type = Type::INT;
	break;
    case Token::TYPE_SHORT :
	t.type = Type::SHORT;
	break;
    case Token::TYPE_BYTE :
	t.type = Type::BYTE;
	break;
    case Token::TYPE_NONE :
	if (!allow_none && !t.ptr)
	    throw Exception(Exception::NONE_VARIABLE, *m_token);
	t.type = Type::NONE;
	break;
    default:
	throw Exception(Exception::UNEXPECTED_TOKEN, *m_token);
	break;
    }
    advance();
}

void Compiler::addLocal(Variable& var, int& stack_offset, bool is_param)
{
    stack_offset += 4; /* Everything aligned to 32 bits. */
    
    var.loc = string(itoa((is_param ? 1 : -1) * (stack_offset + 4*is_param),
			  10)) + "(%ebp)";
    
    if (!var.name.empty())
	if (!m_locals.insert(pair<string, Variable>(var.name, var)).second)
	    throw Exception(Exception::DUPLICATED_LOCAL, *m_token);
}

void Compiler::readLocals(Function& f, bool is_param)
{
    f.locals_size = 0;

    while (m_token->isType() ||
	   (is_param && m_token->getType() == Token::ETCETERA))
    {
	Variable var;

	if (m_token->isType()) 
	    readType(var.type, false);
	else if (m_token->getType() == Token::ETCETERA) {
	    f.params_size = -f.params_size;
	    advance();
	    break;
	}
	
	if (is_param)
	    f.par_types.push_back(var.type);

	if (m_token->getType() == Token::USER_SYMBOL) {
	    var.name = m_token->getString();
	    advance();
	} else if (!(is_param &&
		     (m_token->getType() != Token::END_BLOCK ||
		      m_token->getType() != Token::ETCETERA))) {
	    throw Exception(Exception::UNEXPECTED_TOKEN, *m_token);
	}
	
	addLocal(var, is_param ? f.params_size : f.locals_size, is_param);
    }

    if (is_param)
	advance();
}

void Compiler::declareFunction(Function& f)
{
    if (m_globals.find(f.name) != m_globals.end())
	throw Exception(Exception::DUPLICATED_SYMBOL, *m_token);
    
    map<string, Function>::iterator itf2 = m_functions.find(f.name);
    if (itf2 != m_functions.end()) {
	Function& f2 = itf2->second;
	
	if (f2.defined)
	    throw Exception(Exception::DUPLICATED_FUNCTION, *m_token);

	if (f.ret_type != f2.ret_type || f.params_size != f2.params_size)
	    throw Exception(Exception::FUNCTION_TYPES_MISMATCH, *m_token);

	list<Type>::iterator i1, i2;
	for (i1 = f.par_types.begin(), i2 = f2.par_types.begin();
	     i1 != f.par_types.end() && i2 != f2.par_types.end();
	     ++i1, ++i2)
	    if (*i1 != *i2)
		throw Exception(Exception::FUNCTION_TYPES_MISMATCH, *m_token);
	if (i1 != f.par_types.end() || i2 != f2.par_types.end())
	    throw Exception(Exception::FUNCTION_TYPES_MISMATCH, *m_token);
    } else {
	m_functions[f.name] = f;
    }    
}

void Compiler::outputFunctionHeader(const Function& f)
{
    *m_out << ".globl " << f.loc << endl;
    *m_out << f.loc << ":\n";
	    
    *m_out << "\tpushl %ebp\n"
	"\tmovl %esp,%ebp\n";

    int stack_size = m_stack_var + m_stack_call + 4*m_tmp_user.size();
    for (int i = REG_EBX; i < MAX_REG; ++i)
	if (m_used_reg[i]) stack_size += 4;
    
    if (stack_size)
	*m_out << "\tsubl $" << stack_size << ",%esp\n\n";   

    stack_size = m_stack_var + 4*m_tmp_user.size() + 4;
    for (int i = REG_EBX; i < MAX_REG; ++i)
	if (m_used_reg[i]) {
	    *m_out << "\tmovl " << REG_NAME[i][4] << ',' <<
		-stack_size << "(%ebp)\n";
	    stack_size += 4;
	}
}

void Compiler::outputFunctionFooter()
{
    int stack_size = m_stack_var + 4*m_tmp_user.size() + 4;
    
    for (int i = REG_EBX; i < MAX_REG; ++i)
	if (m_used_reg[i]) {
	    *m_out << "\tmovl " << -stack_size << "(%ebp),"
		   << REG_NAME[i][4] << endl;
	    stack_size += 4;
	}
    
    *m_out << "\n"
	"\tmovl %ebp,%esp\n"
	"\tpopl %ebp\n";
    
    *m_out << "ret\n\n";
}

void Compiler::defineFunction()
{
    Function f;
    
    m_locals.clear();

    advance();
    if (m_token->getType() != Token::USER_SYMBOL)
	throw Exception(Exception::UNEXPECTED_TOKEN, *m_token);
    f.name = m_token->getString();

    if (m_globals.find(f.name) != m_globals.end())
	throw Exception(Exception::DUPLICATED_SYMBOL, *m_token);
    
    f.loc = labelizeString(f.name);
    advance();
    
    readType(f.ret_type, true);
    m_ret_type = f.ret_type;    
    readLocals(f, true);
    declareFunction(f);
    if (m_token->getType() == Token::END_BLOCK) {
	advance();
	return;
    }

    m_functions[f.name].defined = true;
    readLocals(f, false);

    m_tmp_user.clear();
    m_code_blocks.clear();
    m_code_blocks.push_back(new ostringstream);
    memset(&m_used_reg, 0, sizeof(bool) * MAX_REG);
    m_stack_var = f.locals_size;
    m_stack_call = 0;
    
    sink() << endl;
    while(expandSentence());
    
    outputFunctionHeader(f);
    for (std::list<ostringstream*>::iterator i = m_code_blocks.begin();
	 i != m_code_blocks.end();
	 ++i) {
	string str = (**i).str();
	if (!str.empty()) {
	    *m_out << str;
	    outputFunctionFooter();
	}
	delete *i;
    }
    m_code_blocks.clear();
}

void Compiler::declareGlobal()
{
    Variable variable;

    readType(variable.type, false);

    if (m_token->getType() != Token::USER_SYMBOL)
	throw Exception(Exception::UNEXPECTED_TOKEN, *m_token);
    
    variable.name = m_token->getString();
    variable.loc = labelizeString(variable.name);

    advance();
    
    if (m_functions.find(variable.name) != m_functions.end())
	throw Exception(Exception::DUPLICATED_SYMBOL, *m_token);
    
    if (!m_globals.insert(pair<string, Variable>(variable.name, variable)).second)
	throw Exception(Exception::DUPLICATED_SYMBOL, *m_token);
}

void Compiler::startCode()
{
    *m_out << ".text\n";
}

void Compiler::startData()
{
    *m_out << ".data\n";
}

void Compiler::outputGlobalVariables()
{
    map<string, Variable>::iterator i;
    for (i = m_globals.begin(); i != m_globals.end(); ++i) {
	*m_out << ".comm " << i->second.loc
	       << ' ' << i->second.type.size()
	       << ',' << i->second.type.size() << endl;
/*
 *m_out << i->second.name << ":\n";
	
 if (i->second.type.ptr)
 *m_out << "\t.long 0\n";
 else
 switch(i->second.type.type) {
 case Type::BYTE:
 *m_out << "\t.byte 0\n";
 break;
 case Type::SHORT:
 *m_out << "\t.word 0\n";
 break;
 case Type::INT:
 *m_out << "\t.long 0\n";
 default:
 *m_out << "\tERROR\n";
 break;
 }
*/
    }
}

void Compiler::outputStringConstants()
{
    map<string, string>::iterator i;
    for (i = m_str_constants.begin(); i != m_str_constants.end(); ++i)
	*m_out << i->second << ":\n"
	       << "\t.string " << i->first << endl;
}

void Compiler::outputStartFunction()
{
    if (m_functions.find("main") != m_functions.end())
	*m_out << "\n.globl _start\n"
	       << "_start:\n"
	       << "\tcall main\n"
	       << "\tmovl %eax,%ebx\n"
	       << "\tmovl $1,%eax\n"
	       << "\tint $0x80\n\n\n";
}

void Compiler::run(const Lexer& l, const char* fout, bool start)
{
    if (fout) {
	m_out = new ofstream(fout);
	if (m_out->fail()) {
	    delete m_out;
	    throw Exception(Exception::FILE_ERROR);
	}
    } else
	m_out = &cout;
    
    m_token = l.begin();
    m_endtoken = l.end();
    m_label_count = 0;

    startCode();
    while (m_token->getType() != Token::END) {
	switch((*m_token).getType()) {
	case Token::FUNC:
	    defineFunction();
	    break;
	case Token::END:
	    break;
	default:
	    if ((*m_token).isType()) {
		declareGlobal();
	    } else
		throw Exception(Exception::UNEXPECTED_TOKEN, *m_token);
	    break;
	}
    }

    if (start)
	outputStartFunction();
    
    startData();
    outputGlobalVariables();
    outputStringConstants();
    
    if (fout)
	delete m_out;
    
    m_out = NULL;
}
