/***************************************************************************
 *                                                                         *
 *   PLIS COMPILER                                                         *
 *   =============                                                         *
 *                                                                         *
 *   Copyright (C) Juan Pedro Bolivar Puente 2007                          *
 *                                                                         *
 *   This program is free software: you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 *                                                                         *
 ***************************************************************************/

#ifndef COMPILER_H
#define COMPILER_H

#include <map>
#include <list>
#include <string>
#include <iostream>
#include <sstream>
#include <vector>

#include "lexer.h"
#include "misc.h"

enum Register {
    REG_NONE,
    REG_EAX,
    REG_EDX,
    REG_EBX,
    REG_ECX,
    MAX_REG
};

enum NormalOp { OP_PLUS, OP_MINUS, OP_AND, OP_OR, OP_XOR, MAX_NORMAL_OP };

enum MathOp { OP_MUL, OP_DIV, OP_MOD, MAX_MATH_OP };

enum ShiftOp { OP_SHIFT_LEFT, OP_SHIFT_RIGHT, MAX_SHIFT_OP };

enum CmpOp { OP_LT, OP_LTEQ, OP_GT, OP_GTEQ, OP_EQ, OP_NEQ, MAX_CMP_OP };

enum IncOp { OP_INC, OP_DEC, MAX_INC_OP };

enum Signed { SIGN_SIGNED = false, SIGN_UNSIGNED = true, MAX_SIGN };

struct Type {
    enum {
	NONE,
	BYTE,
	SHORT,
	INT
    };
	
    int ptr;
    bool unsign;
    bool con;
    int type;

    Type(int t = INT) :
	ptr(0), unsign(0), con(0), type(t) {}
	
    size_t size() const {
	if (ptr) return 4;
	switch(type) {
	case BYTE:  return 1;
	case SHORT: return 2;
	case INT:   return 4; 
	default:    return 0;
	}
    }

    size_t ptrSize() const {
	if (ptr >= 2)
	    return 4;
	else
	    switch (type) {
	    case Type::SHORT:
		return 2;
		break;
	    case Type::INT:
		return 4;
		break;
	    default:
		return 0;
		break;
	    }
    }
    
    bool operator==(const Type& t) const {
	return ptr == t.ptr && type == t.type && con == t.con && unsign == t.unsign;
    }

    bool operator!=(const Type& t) const {
	return ptr != t.ptr || type != t.type || con != t.con || unsign != t.unsign;
    }

    bool isNone() const {
	return !ptr && type == NONE;
    }
};
	
struct Variable {
    Type type;
    std::string loc;
    std::string name;
};

struct Function {
    Type ret_type;
    std::list<Type> par_types;

    int params_size;
    int locals_size;
    bool defined;
    std::string name;
    std::string loc;
    
    Function() :
	params_size(0), locals_size(0), defined(false) {}
};

struct Jump {
    std::string name;
    bool follows;
};

struct Value {
    enum Location {
	NONE,
	MEMORY_STACK,
	MEMORY_GLOBAL,
	MEMORY_REF,
	CONSTANT,
	PTR
    };
	
    Type type;
    
    int reg;
    int tmp;

    Location loc_type;
    union {
	std::string* mem;
	std::string* ptr;
	int val;
    } loc;
	
    Value(Type t = Type()) :
	type(t), reg(REG_NONE), tmp(-1), loc_type(NONE) {}
		
    Value(Type t, std::string* m, Location loc_type) :
	type(t), reg(REG_NONE), tmp(-1), loc_type(loc_type) {
	loc.mem = m;
    }

    Value(Type t,  int v) :
	type(t), reg(REG_NONE), tmp(-1), loc_type(CONSTANT) {
	loc.val = v;
    }

    void clearLocation() {
	loc_type = NONE;
    }
	
    bool memoryOnly() const {
	return (loc_type == MEMORY_REF || reg == REG_NONE) && isMemory();
    }

    bool constantOnly() const {
	return reg == REG_NONE &&
	    (loc_type == CONSTANT || loc_type == PTR);
    }

    bool isPointer() const {
	return type.ptr;
    }
    
    bool inRegister() const {
	return loc_type != MEMORY_REF && reg != REG_NONE;
    }

    bool inTemporal() const {
	return tmp >= 0;
    }

    bool isConstant() const {
	return (loc_type == CONSTANT || loc_type == PTR);	
    }
    
    bool isNumber() const {
	return loc_type == CONSTANT;
    }

    bool isMemory() const {
	return loc_type == MEMORY_STACK || loc_type == MEMORY_GLOBAL
	    || loc_type == MEMORY_REF;
    }

    size_t size() const {
	return type.size();
    }
};

class Compiler
{
    std::map<std::string, Variable> m_globals;
    std::map<std::string, Variable> m_locals;
    std::map<std::string, Function> m_functions;
    std::map<std::string, std::string> m_str_constants;
    
    Lexer::ConstIterator m_token;
    Lexer::ConstIterator m_endtoken;

    std::list<std::ostringstream*> m_code_blocks;
    std::ostream* m_out;

    std::list<std::string> m_tmp_str;
    std::string m_loop_tag;
    std::string m_stop_tag;
    int m_label_count;
    Type m_ret_type;
    int m_stack_var;
    int m_stack_call;
    Value* m_reg_user[MAX_REG];
    std::vector<Value*> m_tmp_user;
    bool m_used_reg[MAX_REG];

    std::string getNewLabel();
    Jump getNewJump(bool follows);
    
    std::ostream& sink();

    std::string labelizeString(const std::string& s);
    std::string outputMemoryValue(Value* val);
    std::string outputValue(Value* val);
    void allocTemporal(Value* val);
    int evacuateRegister(int exclude);
    void toRegister(Value* val);
    void toRegisterForce(Value* val, Register reg);
    void toRegisterExclude(Value* val, Register reg);
    void changeRegister(Value* val);
    void clearRegister(Value* val);
    void clearTemporal(Value* val);
    void clear(Value* val);
    void resetTemporals();
    void resetRegisters();
    void setRegister(Value* val, Register reg);
    void freeRegister(Register reg);
    void freeRegisterEaxEdx();
    int findRegister(Value* val);
    
    void maximizeTypes(Value* val1, Value* val2, bool one_ptr);
    void castValue(Value* val, const Type& t);

    void expandParam(Type t, std::list<Value*>& params);
    Value* expressionDereference(bool ret);
    Value* expressionIndex(bool ret);
    Value* expressionIndirection(bool ret);
    Value* expressionConditional(bool ret);
    Value* expressionShiftOp(bool ret, ShiftOp op);
    Value* expressionNormalOp(bool ret, NormalOp op);
    Value* expressionMathOp(bool ret, MathOp op);
    Value* expressionIncOp(bool ret, IncOp op, bool post);
    Value* expressionSet(bool ret);
    Value* expressionNumber(bool ret);
    Value* expressionString(bool ret);
    Value* expressionCasting(bool ret);
    Value* expressionSize(bool ret);
    Value* expressionParam(bool ret);
    Value* tryExpressionFunction(bool ret);
    Value* tryExpressionVariable(bool ret);
    Value* expression(bool ret);

    void conditionalOr(Jump& truejmp, Jump& falsejmp);
    void conditionalAnd(Jump& truejmp, Jump& falsejmp);
    void conditionalNot(Jump& truejmp, Jump& falsejmp);
    void conditionalXor(Jump& truejmp, Jump& falsejmp);
    void conditionalArithmetic(Jump& truejmp, Jump& falsejmp, CmpOp op);
    void conditionalExpression(Jump& truejmp, Jump& falsejmp);
    void conditional(Jump& truejmp, Jump& falsejmp);
    
    void structIf();
    void structWhile();
    void structRepeat();
    void structFor();

    void expandFlowLoop();
    void expandFlowStop();
    void expandReturn();
    bool expandSentence();
    
    void declareGlobal();
    void readType(Type& t, bool allow_none);
    void addLocal(Variable& var, int& stack_offset, bool is_param);
    void readLocals(Function& f, bool is_param);
    void declareFunction(Function& f);
    void defineFunction();
    void outputFunctionHeader(const Function& f);
    void outputFunctionFooter();
    
    void startCode();
    void startData();
    void outputGlobalVariables();
    void outputStringConstants();
    void outputStartFunction();
    
    void advance() { ++m_token; }

public:
    Compiler() :
	m_label_count(0), m_stack_var(0), m_stack_call(0) {
	memset(m_reg_user, 0, sizeof(Value*) * MAX_REG);
    }
    
    void run(const Lexer& l, const char* out, bool start);
    
    class Exception : public virtual std::exception
    {
	int m_id;
	Token m_tk; /* Esto tiene un string, hum... */
    public:
	enum {
	    FILE_ERROR,
	    UNEXPECTED_TOKEN,
	    DUPLICATED_LOCAL,
	    DUPLICATED_SYMBOL,
	    DUPLICATED_FUNCTION,
	    FUNCTION_TYPES_MISMATCH,
	    FLOW_MODIFIER_OUTSIDE_LOOP,
	    UNIMPLEMENTED_FEATURE,
	    INVALID_PARAM_TYPE,
	    INVALID_INDEX,
	    INVALID_DEREFERENCE,
	    INVALID_INDIRECTION,
	    NO_RETURN_VALUE,
	    NONE_VARIABLE,
	    UNDEFINED_SYMBOL,
	    ASSIGN_TEMPORAL,
	    REMOVE_CONSTNESS,
	    NO_ENOUGH_REG,
	    N_ERROR
	};

	Exception (int id) :
	    m_id(id) {};
	
	Exception(int id, const Token& tk) :
	    m_id(id), m_tk(tk) {};

	virtual ~Exception() throw() {}
	
	int getID() const {
	    return m_id;
	};
	
	const Token& getToken() const {
	    return m_tk;
	};
	
	virtual const char* what() const throw() {
	    switch (m_id) {
	    case NO_ENOUGH_REG:
		return "All machine registers are used."
		    "This is a compiler limitation tha should be solved in the future.";
	    case ASSIGN_TEMPORAL:
		return "Cannot assign to temporal value";
	    case FUNCTION_TYPES_MISMATCH:
		return "Types in function declaration and definition do not match";
	    case FLOW_MODIFIER_OUTSIDE_LOOP:
		return "Loop flow manipulators must be used inside a loop";
	    case DUPLICATED_LOCAL:
		return (std::string("Duplicated local variable ") + m_tk.getString()).c_str();
	    case DUPLICATED_SYMBOL:
		return (std::string("Duplicated global symbol declaration ") + m_tk.getString()).c_str();
	    case DUPLICATED_FUNCTION:
		return (std::string("Duplicated function definition ") + m_tk.getString()).c_str();
	    case UNDEFINED_SYMBOL:
		return (std::string("Undefined symbol ") + m_tk.getString()).c_str();
	    case INVALID_PARAM_TYPE:
		return "Invalid expression parameter type";
	    case NO_RETURN_VALUE:
		return "Can not use the return value of a function returning 'none'.";
	    case INVALID_INDEX:
		return "Index must be an integer value";
	    case INVALID_DEREFERENCE:
		return "Can not dereference a non-pointer type";
	    case INVALID_INDIRECTION:
		return "Only memory locations can be indirected";
	    case UNIMPLEMENTED_FEATURE:
		return (std::string("Feature not implemented: ") + m_tk.getString()).c_str();
	    case FILE_ERROR:
		return "Could not open output file";
	    case REMOVE_CONSTNESS:
		return "Trying to remove type constness";
	    case NONE_VARIABLE:
		return "'none' is not a valid variable type";
	    case UNEXPECTED_TOKEN:
		return (std::string("Unexpected token ") + m_tk.getString()).c_str(); 
	    default:
		return "Unknown exception.";
	    }
	}
    };
};

#define TYPE_MAX_SIZE 5
#define MAX_BOOL 2
#define MAX_SIGN 2

extern const char* REG_NAME[MAX_REG][TYPE_MAX_SIZE];
extern const char SIZE_MOD[TYPE_MAX_SIZE];


#endif /* COMPILER_H */
