/***************************************************************************
 *                                                                         *
 *   PLIS COMPILER                                                         *
 *   =============                                                         *
 *                                                                         *
 *   Copyright (C) Juan Pedro Bolivar Puente 2007                          *
 *                                                                         *
 *   This program is free software: you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 *                                                                         *
 ***************************************************************************/

#include "compiler.h"

using namespace std;

const char* CMP_OP_NAME[MAX_BOOL][MAX_SIGN][MAX_CMP_OP] = {
    {{"jl", "jle", "jg", "jge", "je", "jne"},
     {"jb", "jbe", "ja", "jae", "je", "jne"}},
    {{"jnl", "jnle", "jng", "jnge", "jne", "je"},
     {"jnb", "jnbe", "jna", "jnae", "jne", "je"}}
};

void Compiler::conditionalArithmetic(Jump& truejmp, Jump& falsejmp, CmpOp op)
{
    advance();
    
    Value* val1 = expression(true);
    Value* val2 = expression(true);

    maximizeTypes(val1, val2, false);

    if (val1->isNumber() && val2->isNumber()) {
	bool op_res = 0;
	
	switch(op) {
	case OP_LT:
	    op_res = val1->loc.val < val2->loc.val;
	    break;
	case OP_LTEQ:
	    op_res = val1->loc.val <= val2->loc.val;
	    break;
	case OP_GT:
	    op_res = val1->loc.val > val2->loc.val;
	    break;
	case OP_GTEQ:
	    op_res = val1->loc.val >= val2->loc.val;
	    break;
	case OP_EQ:
	    op_res = val1->loc.val == val2->loc.val;
	    break;
	case OP_NEQ:
	    op_res = val1->loc.val != val2->loc.val;
	    break;
	default:
	    break;
	}

	if (op_res) {
	    if(!truejmp.follows)
		sink() << "\tjmp " << truejmp.name << endl;
	} else if (!falsejmp.follows) {
	    if (!falsejmp.follows)
		sink() << "\tjmp " << falsejmp.name << endl;
	}
	    
	clear(val1);
	clear(val2);
	delete val1;
	delete val2;

	return;
    }

    Jump tjmp = truejmp;
    Jump fjmp = falsejmp;
    bool negate = false;
    
    if (truejmp.follows) {
	if (falsejmp.follows) {
	    clear(val1);
	    clear(val2);
	    delete val1;
	    delete val2;
	    return;
	} else {
	    negate = true;
	    swap(tjmp, fjmp);
	}
    }

    if (!val1->inRegister() && !val2->inRegister())
	toRegister(val1);
    
    sink() << "\tcmp" << SIZE_MOD[val1->size()] << ' '
	   << outputValue(val2) << ',' << outputValue(val1) << endl;
    sink() << '\t' << CMP_OP_NAME[negate][val1->type.unsign][op] << ' '
	   << tjmp.name << endl;

    if (!fjmp.follows)
	sink() << "\tjmp " << fjmp.name << endl;
    
    clear(val1);
    clear(val2);
    delete val1;
    delete val2;
}

void Compiler::conditionalExpression(Jump& truejmp, Jump& falsejmp)
{
    Value* val = expression(true);

    if (val->isNumber()) {
	if (val->loc.val) {
	    if (!truejmp.follows)
		sink() << "\tjmp " << truejmp.name << endl;
	} else if (!falsejmp.follows)
	    sink() << "\tjmp" << falsejmp.name << endl;

	clear(val);
	delete val;
	return;
    }

    sink() << "\tcmp" << SIZE_MOD[val->size()] << ' '
	   << "$0," << outputValue(val) << endl; 

    if (!truejmp.follows) {
	sink() << "\tjne " << truejmp.name << endl;
	if (!falsejmp.follows)
	    sink() << "\tjmp " << falsejmp.name << endl;
    } else if (!falsejmp.follows)
	sink() << "\tje " << falsejmp.name << endl;
	
    clear(val);
    delete val;
}

void Compiler::conditionalNot(Jump& truejmp, Jump& falsejmp)
{
    advance();

    Jump fjmp = truejmp;
    Jump tjmp = falsejmp;
    
    conditional(tjmp, fjmp);
}

void Compiler::conditionalOr(Jump& truejmp, Jump& falsejmp)
{
    advance();

    Jump tmp = getNewJump(true);
    bool old_follows = truejmp.follows;
    truejmp.follows = false;

    conditional(truejmp, tmp);
    
    sink() << tmp.name << ":\n";
    truejmp.follows = old_follows;
    
    conditional(truejmp, falsejmp);
}

void Compiler::conditionalAnd(Jump& truejmp, Jump& falsejmp)
{
    advance();

    Jump tmp = getNewJump(true);
    bool old_follows = falsejmp.follows;
    falsejmp.follows = false;
    
    conditional(tmp, falsejmp);

    sink() << tmp.name << ":\n";
    falsejmp.follows = old_follows;

    conditional(truejmp, falsejmp);
}

void Compiler::conditionalXor(Jump& truejmp, Jump& falsejmp)
{
    throw Exception(Exception::UNIMPLEMENTED_FEATURE, *m_token);
}

void Compiler::conditional(Jump& truejmp, Jump& falsejmp)
{
    switch(m_token->getType()) {
    case Token::OPERATOR_LT:
	conditionalArithmetic(truejmp, falsejmp, OP_LT);
	break;
    case Token::OPERATOR_LTEQ:
	conditionalArithmetic(truejmp, falsejmp, OP_LTEQ);
	break;
    case Token::OPERATOR_GT:
	conditionalArithmetic(truejmp, falsejmp, OP_GT);
	break;
    case Token::OPERATOR_GTEQ:
	conditionalArithmetic(truejmp, falsejmp, OP_GTEQ);
	break;
    case Token::OPERATOR_EQ:
	conditionalArithmetic(truejmp, falsejmp, OP_EQ);
	break;
    case Token::OPERATOR_NEQ:
	conditionalArithmetic(truejmp, falsejmp, OP_NEQ);
	break;
    case Token::OPERATOR_NOT:
	conditionalNot(truejmp, falsejmp);
	break;
    case Token::OPERATOR_AND:
	conditionalAnd(truejmp, falsejmp);
	break;
    case Token::OPERATOR_OR:
	conditionalOr(truejmp, falsejmp);
	break;
    case Token::OPERATOR_XOR:
	conditionalXor(truejmp, falsejmp);
	break;
    default:
	conditionalExpression(truejmp, falsejmp);
	break;
    }
}

