/***************************************************************************
 *                                                                         *
 *   PLIS COMPILER                                                         *
 *   =============                                                         *
 *                                                                         *
 *   Copyright (C) Juan Pedro Bolivar Puente 2007                          *
 *                                                                         *
 *   This program is free software: you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 *                                                                         *
 ***************************************************************************/

#include "compiler.h"

using namespace std;

const char* NORMAL_OP_NAME[MAX_NORMAL_OP] = {
    "add", "sub", "and", "or", "xor"
};

const char* MATH_OP_NAME[MAX_SIGN][MAX_MATH_OP] = {
    {"imul", "idiv", "idiv"},
    {"mul", "div", "div"}
};

const char* CONVERSION_OP[TYPE_MAX_SIZE] = {
    "ERROR", "cbw", "cwd", "ERROR", "cdq"
};

const char* SHIFT_OP_NAME[MAX_SHIFT_OP] = {
    "shl", "shr"
};

const char* INC_OP_NAME[MAX_INC_OP] = {
    "add", "sub"
};

Value* Compiler::expressionNormalOp(bool ret, NormalOp op)
{
    advance();

    Value* val1 = expression(ret);
    Value* val2 = expression(ret);
    
    if (!ret) {
	if (val1) clear(val1);
	if (val2) clear(val2);
	delete val1;
	delete val2;
	return NULL;
    }

    maximizeTypes(val1, val2, true);

    if (op != OP_MINUS && /* Make sure the op is commutative */
	((!val1->inRegister() && val2->inRegister()) ||
	 val1->isPointer()))
	swap(val1, val2);
    
    if (val1->isPointer())
	throw Exception(Exception::INVALID_PARAM_TYPE, *m_token);
    
    if (val1->isNumber() && val2->isNumber()) {
	switch (op) {
	case OP_PLUS:
	    if (val2->isPointer())
	    val1->loc.val = val1->loc.val + val2->loc.val;
	    break;
	case OP_MINUS:
	    val1->loc.val = val1->loc.val - val2->loc.val;
	    break;
	case OP_AND:
	    val1->loc.val = val1->loc.val & val2->loc.val;
	    break;
	case OP_OR:
	    val1->loc.val = val1->loc.val | val2->loc.val;
	    break;
	case OP_XOR:
	    val1->loc.val = val1->loc.val ^ val2->loc.val;
	    break;
	default:
	    break;
	}
	clear(val1);
	clear(val2);
	delete val2;
	return val1;
    }
    
    toRegister(val1);
    
    if (val2->isPointer() &&
	val2->type.ptr < 2 &&
	val2->type.type != Type::BYTE)
    {
	sink() << "\tshl" << SIZE_MOD[val1->size()] << ' ';

	if (val2->type.ptr >= 2)
	    sink() << "$2";
	else switch (val2->type.type) {
	    case Type::SHORT:
		sink() << "$1";
		break;
	    case Type::INT:
		sink() << "$2";
		break;
	    default:
		sink() << "ERROR";
		break;
	    }
	sink() << ',' << outputValue(val1) << endl;
    }
    
    sink() << '\t' << NORMAL_OP_NAME[op] << SIZE_MOD[val1->size()]
	   << ' ' << outputValue(val2) << ',' << outputValue(val1) << endl;

    if (val2->isPointer())
	val1->type = val2->type;
    
    val1->clearLocation();
    clear(val2);
    delete val2;
    return val1;
}

Value* Compiler::expressionMathOp(bool ret, MathOp op)
{
    advance();
    
    Value* val1 = expression(ret);
    Value* val2 = expression(ret);
    
    if (!ret) {
	if (val1) clear(val1);
	if (val2) clear(val2);
	delete val1;
	delete val2;
	return NULL;
    }

    if (val1->isPointer() || val2->isPointer())
	throw Exception(Exception::INVALID_PARAM_TYPE, *m_token);
        
    maximizeTypes(val1, val2, true); /* TODO: Make maximizeTypes alloc the correct reg. */
    
    if (val1->isNumber() && val2->isNumber()) {
	switch (op) {
	case OP_MUL:
	    val1->loc.val = val1->loc.val * val2->loc.val;
	    break;
	case OP_DIV:
	    val1->loc.val = val1->loc.val / val2->loc.val;
	    break;
	case OP_MOD:
	    val1->loc.val = val1->loc.val % val2->loc.val;
	    break;
	default:
	    break;
	}
	clear(val1);
	clear(val2);
	delete val2;
	return val1;
    }

    if (op == OP_MUL &&
	val2->constantOnly())
	swap(val1, val2);
    
    toRegisterForce(val1, REG_EAX);
    if (val2->constantOnly())
	toRegisterExclude(val2, REG_EDX);
    freeRegister(REG_EDX);

    if (op == OP_DIV || op == OP_MOD)
	sink() << '\t' << CONVERSION_OP[val1->size()] << endl;
    
    sink() << '\t' << MATH_OP_NAME[val2->type.unsign][op]
	   << SIZE_MOD[val2->size()]
	   << ' ' << outputValue(val2) << endl;

    if (op == OP_MOD) {
	if (val1->size() == 1);
	/* TODO! */
	else setRegister(val1, REG_EDX);
    } else setRegister(val1, REG_EAX);
    
    val1->clearLocation();
    
    clear(val2);
    delete val2;
    return val1;
}

Value* Compiler::expressionShiftOp(bool ret, ShiftOp op)
{
    advance();
    
    Value* val1 = expression(ret);
    Value* val2 = expression(ret);

    /* If no return value is needed, just evaluate parameters. */
    if (!ret) {
	if (val1) clear(val1);
	if (val2) clear(val2);
	delete val1;
	delete val2;
	return NULL;
    }

    if (val1->isPointer() || val2->isPointer())
	throw Exception(Exception::INVALID_PARAM_TYPE, *m_token);
    
    if (val1->isNumber() && val2->isNumber()) {
	switch(op) {
	case OP_SHIFT_LEFT:
	    val1->loc.val = val1->loc.val << val2->loc.val;
	    break;
	case OP_SHIFT_RIGHT:
	    val1->loc.val = val1->loc.val >> val2->loc.val;
	    break;
	default:
	    break;
	}
	clear(val1);
	clear(val2);
	delete val2;
	return val1;
    }

    toRegister(val1);
    if (!val2->isNumber())
	toRegisterForce(val2, REG_ECX);

    sink() << "\t" << SHIFT_OP_NAME[op] << SIZE_MOD[val1->size()] << ' ';
    if (val2->reg == REG_NONE)
	sink() << '$' << val2->loc.val;
    else
	sink() << "%cl";
    sink() << ',' << outputValue(val1) << endl;
    
    clear(val2);
    return val1;
    
}

Value* Compiler::expressionSet(bool ret)
{
    advance();
    
    Value* val1 = expression(true);
    Value* val2 = expression(true);
    
    if (!val1->isMemory())
	throw Exception(Exception::ASSIGN_TEMPORAL, *m_token);

    castValue(val2, val1->type);
    
    if (val2->memoryOnly())
	toRegister(val2);
    
    sink() << '\t' << "mov" << SIZE_MOD[val1->size()]
	   << ' ' << outputValue(val2) << ','
	   << outputMemoryValue(val1) << endl;

    Register reg = (Register)val2->reg;
    clear(val2);
    setRegister(val1, reg);
    delete val2;
    return val1;
}

void Compiler::expandParam(Type t, list<Value*>& params)
{
    Value* val = expression(true);

    if (!t.isNone())
	castValue(val, t);
	    
    params.push_back(val);
}

Value* Compiler::tryExpressionFunction(bool ret)
{   
    Value* val = NULL;
    map<string, Function>::iterator it;
    
    it = m_functions.find(m_token->getString());
    if (it != m_functions.end()) {
	advance();
	Function& f = it->second;
	list<Type>::iterator it;
	list<Value*> params;

	if (ret && f.ret_type.isNone())
	    throw Exception(Exception::NO_RETURN_VALUE, *m_token);
	
	for (it = f.par_types.begin(); it != f.par_types.end(); ++it)
	    expandParam(*it, params);

	if (f.params_size < 0) {
	    while (m_token->getType() != Token::END_BLOCK)
		expandParam(Type(Type::NONE), params);
	    advance();
	}

	int offset = 0;
	for (list<Value*>::iterator it = params.begin(); it != params.end(); ++it) {
	    if (!(*it)->isConstant())
		toRegister(*it);
	    sink() << "\tmov" << SIZE_MOD[(*it)->size()] << ' '
		   << outputValue(*it) << ',' << offset << "(%esp)" << endl;
	    clear(*it);
	    delete *it;
	    offset += 4;
	}

	if ((signed)(4 * params.size()) > m_stack_call)
	    m_stack_call = 4 * params.size();
	
	freeRegisterEaxEdx();

	sink() << "\tcall " << f.loc << endl;

	val = new Value(f.ret_type);
	setRegister(val, REG_EAX);
    }
    
    return val;
}

Value* Compiler::tryExpressionVariable(bool ret)
{
    map<string, Variable>::iterator it;

    Value::Location loc_type;
    
    it = m_locals.find(m_token->getString());
    if (it == m_locals.end()) {
	it = m_globals.find(m_token->getString());
	if (it == m_globals.end())
	    return NULL;
	else
	    loc_type = Value::MEMORY_GLOBAL;
    } else
	loc_type = Value::MEMORY_STACK;

    advance();
    return new Value(it->second.type, &it->second.loc, loc_type);
}

Value* Compiler::expressionNumber(bool ret)
{
    if (!ret) {
	advance();
	return NULL;
    }
    
    int val = atoi(m_token->getString().c_str());
    advance();
    return new Value(Type(), val);
}

Value* Compiler::expressionString(bool ret)
{
    if (!ret) {
	advance();
	return NULL;
    }
    
    map<string,string>::iterator i;
    Value* val;
    
    i = m_str_constants.find(m_token->getString());

    if (i == m_str_constants.end())
	i = m_str_constants.insert(
	    pair<string,string>(m_token->getString(), getNewLabel())).first;

    val = new Value;
    val->type.ptr = 1;
    val->type.type = Type::BYTE;
    val->loc_type = Value::PTR;
    val->loc.ptr = &i->second;
        
    advance();

    return val;
}

Value* Compiler::expressionConditional(bool ret)
{
    advance();

    Value* val = new Value();

    if (ret)
	allocTemporal(val);
    
    Jump setzero = getNewJump(ret);
    Jump setone = getNewJump(false);

    conditional(setone, setzero);

    sink() << setzero.name << ":\n";
    if (ret)
	sink() << "\tmov" << SIZE_MOD[val->size()] << ' '
	       << "$0," << outputValue(val);

    sink() << setone.name << ":\n";
    if (ret)
	sink() << "\tmov" << SIZE_MOD[val->size()] << ' '
	       << "$1," << outputValue(val);
    
    return val;
}

Value* Compiler::expressionIncOp(bool ret, IncOp op, bool post)
{
    advance();
    
    Value* val = expression(true);
    
    if (!val->isMemory())
	throw Exception(Exception::ASSIGN_TEMPORAL, *m_token);

    if (post && ret)
	toRegister(val);
    
    sink() << '\t' << INC_OP_NAME[op] << SIZE_MOD[val->size()]
	   << " $" << (val->type.ptr ? val->size() : 1)
	   << ',' << outputMemoryValue(val) << endl;
    
    if (post)
	val->clearLocation();

    return val;
}

Value* Compiler::expressionCasting(bool ret)
{
    Type new_type;
    readType(new_type, false);
    
    Value* val = expression(ret);
    
    if (!ret) {
	if (val) {
	    clear(val);
	    delete val;
	    val = NULL;
	}
    } else {
	castValue(val, new_type);
    }
    
    return val;
}

Value* Compiler::expressionIndex(bool ret)
{
    advance();    
    Value* val1 = expression(ret);
    Value* val2 = expression(ret);
    
    if (!ret) {
	if (val1) clear(val1);
	if (val2) clear(val2);
	delete val1;
	delete val2;
	return NULL;
    }

    if (val1->isPointer())
	throw Exception(Exception::INVALID_INDEX, *m_token);

    if (!val2->isPointer())
	throw Exception(Exception::INVALID_DEREFERENCE, *m_token);

    castValue(val1, Type());
    toRegister(val1);
    toRegister(val2);

    sink() << "\tleal ("
	   << outputValue(val2) << ',' << outputValue(val1) << ','
	   << val2->type.ptrSize() << ")," << outputValue(val2) << endl;
    
    val2->type.ptr--;
    val2->loc_type = Value::MEMORY_REF;

    clear(val1);
    delete val1;
    
    return val2;
}

Value* Compiler::expressionDereference(bool ret)
{
    advance();    
    Value* val = expression(ret);

    if (!ret) {
	if (val) clear(val);
	delete val;
	return NULL;
    }

    if (!val->isPointer())
	throw Exception(Exception::INVALID_DEREFERENCE, *m_token);

    toRegister(val);
    
    val->type.ptr--;
    val->loc_type = Value::MEMORY_REF;
    
    return val;
}

Value* Compiler::expressionIndirection(bool ret)
{
    advance();
    Value* val = expression(ret);

    if (!ret) {
	if (val) clear(val);
	delete val;
	return val;
    }

    if (!val->isMemory())
	throw Exception(Exception::INVALID_INDIRECTION, *m_token);

    if (val->loc_type == Value::MEMORY_STACK) {
	int reg = findRegister(val);
	sink() << "\tleal " << *val->loc.mem
	       << ',' << REG_NAME[reg][4] << endl;
	setRegister(val, (Register)reg);
	val->loc_type = Value::NONE;
    } else if (val->loc_type == Value::MEMORY_GLOBAL) {
	val->loc_type = Value::PTR;
    } else
	val->loc_type = Value::NONE;
    
    val->type.ptr++;

    return val;
}

Value* Compiler::expressionSize(bool ret)
{
    advance();

    int size = 0;

    if (m_token->isType()) {
	Type t;
	readType(t, true);
	size = t.size();
    } else {
	Value* val = expression(true);
	size = val->size();
	clear(val);
	delete val;
    }

    if (ret)
	return new Value(Type(), size);

    return NULL;
}

Value* Compiler::expressionParam(bool ret)
{
    advance();

    Value* val = expression(true);

    if (!ret) {
	clear(val);
	delete val;
	return NULL;
    }

    if (val->isPointer())
	throw Exception(Exception::INVALID_PARAM_TYPE, *m_token);

    if (val->isNumber()) {
	m_tmp_str.push_back(string(itoa(8+val->loc.val*4,10))+"(%ebp)");
	clear(val);
	val->type = Type();
	val->loc_type = Value::MEMORY_STACK;
	val->loc.mem = &m_tmp_str.back();
	return val;
    }
	
    castValue(val, Type());

    toRegister(val);

    int reg = findRegister(val);
    
    sink() << "\tleal "
	   << "8(%ebp," << outputValue(val) << ",4)," << REG_NAME[reg][4] << endl;

    clear(val);
    setRegister(val, (Register)reg);
    
    val->loc_type = Value::MEMORY_REF;
    
    return val;
}

Value* Compiler::expression(bool ret)
{
    Value* val;
    
    switch(m_token->getType()) {
    case Token::USER_SYMBOL:
	if ((val = tryExpressionVariable(ret)) ||
	    (val = tryExpressionFunction(ret)))
	    return val;
	throw Exception(Exception::UNDEFINED_SYMBOL, *m_token);
	break;

    case Token::NUM_CONST:
	return expressionNumber(ret);

    case Token::STRING_CONST:
	return expressionString(ret);
	
    case Token::OPERATOR_PLUS:
	return expressionNormalOp(ret, OP_PLUS);
	
    case Token::OPERATOR_MINUS:
	return expressionNormalOp(ret, OP_MINUS);
    
    case Token::OPERATOR_BITAND:
	return expressionNormalOp(ret, OP_AND);
    
    case Token::OPERATOR_BITOR:
	return expressionNormalOp(ret, OP_OR);
    
    case Token::OPERATOR_BITXOR:
	return expressionNormalOp(ret, OP_XOR);

    case Token::OPERATOR_SET:
	return expressionSet(ret);
	
    case Token::OPERATOR_MUL:
	return expressionMathOp(ret, OP_MUL);

    case Token::OPERATOR_DIV:
	return expressionMathOp(ret, OP_DIV);
	
    case Token::OPERATOR_MOD:
	return expressionMathOp(ret, OP_MOD);

    case Token::OPERATOR_LSHIFT:
	return expressionShiftOp(ret, OP_SHIFT_LEFT);

    case Token::OPERATOR_RSHIFT:
	return expressionShiftOp(ret, OP_SHIFT_RIGHT);
	
    case Token::OPERATOR_PREINC:
	return expressionIncOp(ret, OP_INC, false);
	
    case Token::OPERATOR_POSTINC:
	return expressionIncOp(ret, OP_INC, true);
	
    case Token::OPERATOR_PREDEC:
	return expressionIncOp(ret, OP_DEC, false);
	
    case Token::OPERATOR_POSTDEC: 
	return expressionIncOp(ret, OP_DEC, true);

    case Token::OPERATOR_DEREF:
	return expressionDereference(ret);

    case Token::OPERATOR_DIR:
	return expressionIndirection(ret);

    case Token::OPERATOR_INDEX:
	return expressionIndex(ret);
	
    case Token::OPERATOR_SIZE:
	return expressionSize(ret);

    case Token::OPERATOR_PARAM:
	return expressionParam(ret);
	
    case Token::OPERATOR_IF:
	throw Exception(Exception::UNIMPLEMENTED_FEATURE, *m_token);

    default:
	if (m_token->isConditional())
	    return expressionConditional(ret);
	else if (m_token->isType())
	    return expressionCasting(ret);
	else
	    throw Exception(Exception::UNEXPECTED_TOKEN, *m_token);
	break;
    }

    return NULL;
}

