/***************************************************************************
 *                                                                         *
 *   PLIS COMPILER                                                         *
 *   =============                                                         *
 *                                                                         *
 *   Copyright (C) Juan Pedro Bolivar Puente 2007                          *
 *                                                                         *
 *   This program is free software: you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 *                                                                         *
 ***************************************************************************/

#include "compiler.h"

using namespace std;

std::string Compiler::outputMemoryValue(Value* val)
{
    if (val->loc_type == Value::MEMORY_REF)
	return outputValue(val);

    return *val->loc.mem;
}

std::string Compiler::outputValue(Value* val)
{
    if (val->loc_type != Value::MEMORY_REF && val->reg != REG_NONE)
	return REG_NAME[val->reg][val->type.size()];

    switch(val->loc_type) {
    case Value::MEMORY_STACK:
    case Value::MEMORY_GLOBAL:
	return *val->loc.mem;

    case Value::MEMORY_REF:
	if (val->reg == REG_NONE && val->tmp >= 0) {
	    val->loc_type = Value::NONE;
	    toRegister(val);
	    val->loc_type = Value::MEMORY_REF;
	}
	return string("(") + REG_NAME[val->reg][4] + ")";
	break;
	
    case Value::CONSTANT:
	return std::string("$") + itoa(val->loc.val, 10);

    case Value::PTR:
	return std::string("$") + *val->loc.ptr;
	
    default:
	if (val->tmp >= 0)
	    return std::string(itoa(-(m_stack_var + 4*val->tmp + 4), 10)) + "(%ebp)";
	return "ERROR";
    }
}

void Compiler::resetRegisters()
{
    memset(m_reg_user, 0, sizeof(Value*) * MAX_REG);
}

void Compiler::resetTemporals()
{
    size_t i;
    for (i = 0; i < m_tmp_user.size(); ++i)
	m_tmp_user[i] = NULL;
}

void Compiler::allocTemporal(Value* val)
{
    if (val->tmp < 0) {
	size_t i;
	
	for (i = 0; i < m_tmp_user.size() && m_tmp_user[i] != NULL; i++);

	if (i == m_tmp_user.size())
	    m_tmp_user.push_back(val);
	else
	    m_tmp_user[i] = val;
	
	val->tmp = i;
    }
}

void Compiler::clearTemporal(Value* val)
{
    if (val->tmp >= 0) {
	m_tmp_user[val->tmp] = NULL;
	val->tmp = -1;
    }
}

int Compiler::evacuateRegister(int exclude)
{
    static int reg = REG_EAX;

    while (reg == exclude) 
	if (++reg == MAX_REG)
	    reg = REG_EAX;

    allocTemporal(m_reg_user[reg]);
    
    sink() << "\tmov " << REG_NAME[reg][m_reg_user[reg]->size()] << ','
	   << itoa(-(m_stack_var + 4*m_reg_user[reg]->tmp + 4), 10)
	   << "(%ebp)" << endl;

    m_reg_user[reg]->reg = REG_NONE;
    m_reg_user[reg] = NULL;
    return reg;
}

void Compiler::toRegister(Value* val)
{
    if (!val->inRegister()) {
	int reg;
	
	if (val->reg == REG_NONE)
	    for (reg = REG_EAX; reg < MAX_REG && m_reg_user[reg] != NULL; ++reg);
	else
	    reg = val->reg;
	
	if (reg >= MAX_REG)
	    reg = evacuateRegister(REG_NONE);
	
	sink() << "\tmov" << SIZE_MOD[val->type.size()]
	       << ' ' << outputValue(val)
	       << ',' << REG_NAME[reg][val->type.size()] << endl;

	if (val->loc_type == Value::MEMORY_REF)
	    val->clearLocation();
	
	val->reg = (Register)reg;
	m_reg_user[reg] = val;
    }
}

void Compiler::changeRegister(Value* val)
{
    if (val->reg != REG_NONE) {
	int old = (int)val->reg;
	int reg;
	
	for (reg = REG_EAX;
	     reg < MAX_REG && (reg == old || m_reg_user[reg] != NULL);
	     ++reg);
    
	if (reg >= MAX_REG)
	    reg = evacuateRegister(old);
	
	sink() << "\tmov" << SIZE_MOD[val->size()]
	       << ' ' << REG_NAME[val->reg][val->size()]
	       << ',' << REG_NAME[reg][val->size()] << endl;

	val->reg = (Register)reg;
	m_reg_user[reg] = val;
	m_reg_user[old] = NULL;
    }
}

void Compiler::toRegisterExclude(Value* val, Register reg_ex)
{
    if (val->reg == reg_ex) {
	changeRegister(val);
	if (val->loc_type == Value::MEMORY_REF)
	    val->clearLocation();
    }
    
    if (!val->inRegister()) { 
	int reg;
	
	if (val->reg != REG_NONE)
	    for (reg = REG_EAX; reg < MAX_REG &&
		     (reg == reg_ex || m_reg_user[reg] != NULL); ++reg);
	else
	    reg = val->reg;
	
	if (reg >= MAX_REG)
	    reg = evacuateRegister(reg_ex);
	
	sink() << "\tmov" << SIZE_MOD[val->type.size()]
	       << ' ' << outputValue(val)
	       << ',' << REG_NAME[reg][val->type.size()] << endl;

	if (val->loc_type == Value::MEMORY_REF)
	    val->clearLocation();
		
	val->reg = (Register)reg;
	m_reg_user[reg] = val;
    }
}

void Compiler::toRegisterForce(Value* val, Register reg)
{
    if (val->reg != reg || val->loc_type == Value::MEMORY_REF) {
	if (m_reg_user[reg])
	    changeRegister(m_reg_user[reg]);
	
	sink() << "\tmov" << SIZE_MOD[val->size()]
	       << ' ' << outputValue(val)
	       << ',' << REG_NAME[reg][val->size()] << endl;

	clearRegister(val);
	val->reg = reg;
	m_reg_user[reg] = val;

	if (val->loc_type == Value::MEMORY_REF)
	    val->clearLocation();
    }
}

void Compiler::clearRegister(Value* val)
{
    m_used_reg[val->reg] = true;
    m_reg_user[val->reg] = NULL;
    val->reg = REG_NONE;
}

void Compiler::clear(Value* val)
{
    clearRegister(val);
    clearTemporal(val);
}

void Compiler::freeRegister(Register reg)
{
    if (m_reg_user[reg] != NULL) {
	changeRegister(m_reg_user[reg]);
    }
}

void Compiler::freeRegisterEaxEdx()
{
    bool eax_null = false;
    bool edx_null = false;

    if (!m_reg_user[REG_EAX]) {
	eax_null = true;
	m_reg_user[REG_EAX] = (Value*)1;
    }
    if (!m_reg_user[REG_EDX]) {
	edx_null = true;
	m_reg_user[REG_EDX] = (Value*)1;
    }

    if (!eax_null) changeRegister(m_reg_user[REG_EAX]);
    if (!edx_null) changeRegister(m_reg_user[REG_EDX]);

    if (eax_null) m_reg_user[REG_EAX] = NULL;
    if (edx_null) m_reg_user[REG_EDX] = NULL;
}

void Compiler::setRegister(Value* val, Register reg)
{
    m_reg_user[val->reg] = NULL;
    m_reg_user[reg] = val;
    val->reg = reg;
}

int Compiler::findRegister(Value* val)
{
    if (val->reg != REG_NONE)
	return val->reg;

    int reg;
    for (reg = REG_EAX; reg < MAX_REG && m_reg_user[reg] != NULL; ++reg);

    if (reg >= MAX_REG)
	throw Exception(Exception::NO_ENOUGH_REG, *m_token);
	
    return reg;
}

void Compiler::maximizeTypes(Value* val1, Value* val2, bool one_ptr)
{
    if (one_ptr && val1->isPointer() && val2->isPointer())
	throw Exception(Exception::INVALID_PARAM_TYPE, *m_token);
	
    if (val1->size() > val2->size())
	swap(val1, val2);

    if (val1->size() < val2->size()) {
	int reg = findRegister(val1);
	
	sink() << "\tmov" << (val1->type.unsign ? 'z' : 's')
	       << SIZE_MOD[val1->size()] << SIZE_MOD[val2->size()]
	       << ' ' << outputValue(val1) << ',' << REG_NAME[reg][val2->size()] << endl;

	setRegister(val1, (Register)reg);
	
	if (!val2->isPointer())
	    val1->type = val2->type;
	else
	    val1->type.type = Type::INT;
	
    } else if (val1->type.unsign && !val2->type.unsign) {
	if (!val2->isPointer())
	    val1->type = val2->type;
	else
	    val1->type.type = Type::INT;
    }
}

void Compiler::castValue(Value* val, const Type& t)
{
    if (val->size() < t.size() && !val->constantOnly()) {
	int reg = findRegister(val);
	
	sink() << "\tmov" << (val->type.unsign ? 'z' : 's')
	       << SIZE_MOD[val->size()] << SIZE_MOD[t.size()]
	       << ' ' << outputValue(val) << ',' << REG_NAME[reg][t.size()] << endl;

	setRegister(val, (Register)reg);
    }

    val->type = t;
}
