/***************************************************************************
 *                                                                         *
 *   PLIS COMPILER                                                         *
 *   =============                                                         *
 *                                                                         *
 *   Copyright (C) Juan Pedro Bolivar Puente 2007                          *
 *                                                                         *
 *   This program is free software: you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 *                                                                         *
 ***************************************************************************/

#include <fstream>
#include "misc.h"
#include "lexer.h"


int Lexer::m_init = false;
std::map<std::string, Token::Type> Lexer::m_lang;

using namespace std;

Lexer::Lexer()
{
    if (!m_init) {
	m_lang["void"] = Token::VOID_SENTENCE;
	m_lang["return"] = Token::RETURN;
	m_lang["]"] = Token::END_BLOCK;
	m_lang["func"] = Token::FUNC;
	m_lang["..."] = Token::ETCETERA;
	m_lang["+"] = Token::OPERATOR_PLUS;
	m_lang["-"] = Token::OPERATOR_MINUS;
	m_lang["*"] = Token::OPERATOR_MUL;
	m_lang["/"] = Token::OPERATOR_DIV;
	m_lang["%"] = Token::OPERATOR_MOD;
	m_lang[">++"] = Token::OPERATOR_POSTINC;
	m_lang["++>"] = Token::OPERATOR_PREINC;
	m_lang[">--"] = Token::OPERATOR_POSTDEC;
	m_lang["-->"] = Token::OPERATOR_PREDEC;
	m_lang[":="] = Token::OPERATOR_SET;
	m_lang["<"] = Token::OPERATOR_LT;
	m_lang["<="] = Token::OPERATOR_LTEQ;
	m_lang[">"] = Token::OPERATOR_GT;
	m_lang[">="] = Token::OPERATOR_GTEQ;
	m_lang["="] = Token::OPERATOR_EQ;
	m_lang["!="] = Token::OPERATOR_NEQ;
	m_lang["<-"] = Token::OPERATOR_DIR;
	m_lang["->"] = Token::OPERATOR_DEREF;
	m_lang["!"] = Token::OPERATOR_NOT;
	m_lang["&&"] = Token::OPERATOR_AND;
	m_lang["||"] = Token::OPERATOR_OR;
	m_lang["~~"] = Token::OPERATOR_XOR;
	m_lang["&"] = Token::OPERATOR_BITAND;
	m_lang["|"] = Token::OPERATOR_BITOR;
	m_lang["~"] = Token::OPERATOR_BITXOR;
	m_lang[">>"] = Token::OPERATOR_RSHIFT;
	m_lang["<<"] = Token::OPERATOR_LSHIFT;
	m_lang["?"] = Token::OPERATOR_IF;
	m_lang["'"] = Token::OPERATOR_INDEX;
	m_lang["size"] = Token::OPERATOR_SIZE;
	m_lang["param"] = Token::OPERATOR_PARAM;
	m_lang["if"] = Token::STRUCT_IF;
	m_lang["else"] = Token::STRUCT_ELSE;
	m_lang["while"] = Token::STRUCT_WHILE;
	m_lang["repeat"] = Token::STRUCT_REPEAT;
	m_lang["for"] = Token::STRUCT_FOR;
	m_lang["loop"] = Token::FLOW_LOOP;
	m_lang["stop"] = Token::FLOW_STOP;
	m_lang["const"] = Token::TYPE_CONST;
	m_lang["unsigned"] = Token::TYPE_UNSIGNED;
	m_lang["int"] = Token::TYPE_INT;
	m_lang["short"] = Token::TYPE_SHORT;
	m_lang["byte"] = Token::TYPE_BYTE;
	m_lang["ptr"] = Token::TYPE_PTR;
	m_lang["none"] = Token::TYPE_NONE;

	m_init = true;
    }
}

static inline bool isSeparator(char c)
{
    return isspace(c) || c == '(' || c == ')';
}

static inline bool isComment(char c)
{
    return c == '{' || c == ';';
}

void Lexer::next(Token::Context& ctx, std::istream& in, std::string& token)
{
    char c = 0;
    token.clear();
    bool find_separators = true;

    /* Skip separators and comments */
    while (find_separators) {
	while(isSeparator(in.peek())) {
	    c = in.get();
	    ctx.update(c);
	}
	find_separators = false;
	
	if (in.peek() == '{') {
	    int level = 1;
	    ctx.update(in.get());
	    while (level > 0 && in.peek() != EOF) {
		c = in.get();
		ctx.update(c);
		if (c == '{')
		    level++;
		else if (c == '}')
		    level--;
	    }
	    find_separators = true;
	}

	if (in.peek() == ';') {
	    while(in.peek() != EOF && in.peek() != '\n') {
		c = in.get();
		ctx.update(c);
	    }
	    find_separators = true;
	}
    }

    /* String constant */
    if (in.peek() == '"') {
	token += in.get();
	ctx.col++;
	
	char lc = c;
	token += c = in.get();
	while((c != '"' || lc == '\\') && c != EOF) {
	    ctx.update(c);	    
	    lc = c;
	    c = in.get();
	    token += c;
	}
	ctx.col++;
	if (c == EOF)
	    token.clear();
    }

    /* Any other token */
    while (!isSeparator(in.peek()) && !isComment(in.peek()) && !(in.peek() == EOF)) {
	c = in.get();
	token += c;
	ctx.col++;
    }	
    
}

void Lexer::expand(std::string& token)
{
    map<string, string>::iterator i = m_def.find(token);

    if (i != m_def.end()) {
	token = (*i).second;
    }
}

bool Lexer::isNumber(const std::string& str)
{
    size_t i;

    if (!isdigit(str[0]))
	if (!str[0] != '+' && str[0] != '-')
	    return false;
	else if (str.size() < 2)
	    return false;

    for (i = 1; i < str.size(); i++) {
	if (!isdigit(str[1]))
	    return false;
    }

    return true;
}

Token::Type Lexer::findType(const std::string& token)
{
    if (token[0] == '"')
	return Token::STRING_CONST;
    if (isNumber(token))
	return Token::NUM_CONST;

    map<string, Token::Type>::iterator i = m_lang.find(token);
    if (i != m_lang.end())
	return (*i).second;
    
    return Token::USER_SYMBOL;
}

void Lexer::run(const char* file)
{
    string token;
    Token::Context ctx(file, 1, 0);
    ifstream fin;
    
    fin.open(file);
    if (fin.fail())
	throw Exception(Exception::FILE_ERROR, ctx);
    
    do {
	next(ctx, fin, token);
	if (token.size() > 0) {
	    expand(token);

	    if (token == "def") {
		string now, after;
		next(ctx, fin, now);
		next(ctx, fin, after);

		if (now.empty() || after.empty()) {
		    token.clear();
		    throw Exception(Exception::UNEXPECTED_EOF, ctx);
		} else {
		    expand(after);
		    m_def[now] = after;
		}
	    } else if (token == "include") {
		string inc_file;
		next(ctx, fin, inc_file);

		if (inc_file.size() < 3 || inc_file[0] != '"')
		    throw Exception(Exception::INCLUDE_ERROR, ctx);
		else {
		    inc_file.erase(inc_file.begin());
		    inc_file.erase(--inc_file.end());
		    if (inc_file[0] != '/')
			inc_file.insert(0, str_dirname(file));
		    
		    run(inc_file.c_str());
		}
	    } else if (token == "lock") {
		string lock;
		next(ctx, fin, lock);

		if (lock.empty())
		    throw Exception(Exception::UNEXPECTED_EOF, ctx);
		else if (!m_locks.insert(lock).second)
		    do next(ctx, fin, token);
		    while (!token.empty() && token != "unlock");
	    } else if (token != "unlock") {
		m_list.push_back(Token(token, findType(token), ctx));
	    }
	}
    } while (!token.empty());
    fin.close();
}
